# - Try to find BSTools
# Once done this will define
#
#  BSTools_FOUND - system has BSTools
#  BSTools_INCLUDE_DIRS - the BSTools include directory
#  BSTools_LIBRARIES - Link these to use BSTools
#  BSTools_DEFINITIONS - Compiler switches required for using BSTools
#
#  Copyright (c) 2010 Roman Putanowicz <putanowr@l5.pk.edu.pl>
#
#  Redistribution and use is allowed according to the terms of the New
#  BSD license.
#  For details see the accompanying COPYING-CMAKE-SCRIPTS file.
#

if (BSTools_LIBRARIES AND BSTools_INCLUDE_DIRS)
  # in cache already
  set(BSTools_FOUND TRUE)
else (BSTools_LIBRARIES AND BSTools_INCLUDE_DIRS)
  set(BSTools_LIBRARIES "") 
  IF( BSTools_FIND_COMPONENTS )
    FOREACH( component ${BSTools_FIND_COMPONENTS} )
      SET(BSTools_USE_${component} 1 )
    ENDFOREACH( component )
  ENDIF( BSTools_FIND_COMPONENTS )
   
  find_path(BSTools_INCLUDE_DIR NAMES bsfile.h
      PATHS "${BASE_DIR}/include/BSTools" 
            "${BASE_DIR}/../install/include/BSTools"
             NO_DEFAULT_PATH
  )
  find_path(BSTools_INCLUDE_DIR NAMES bsfile.h
  )

  # Order of the libraries is important -- it reflects dependencies
  # between libraries
  FOREACH(module  xgedit raybez bsfile g2blending g1blending bsmesh camera
                  eghole multibs psout pkgeom pknum pkvaria
         )
    if (BSTools_USE_${module})
    MESSAGE(STATUS "checking for library lib${module}.a in ${BASE_DIR_LIB}/BSTools")
      find_library(BSTools_${module}_LIBRARY NAMES ${module} PATHS
                   "${BASE_DIR_LIB}/BSTools"
     "${BASE_DIR_LIB}/../../install/lib/BSTools"
       NO_DEFAULT_PATH
      )
      find_library(BSTools_${module}_LIBRARY NAMES ${module} 
      )

      set(BSTools_LIBRARIES
           ${BSTools_LIBRARIES} ${BSTools_${module}_LIBRARY})
    endif (BSTools_USE_${module})
  ENDFOREACH(module)

  set(BSTools_INCLUDE_DIRS
    ${BSTools_INCLUDE_DIR} CACHE PATH "Path to BSTools headers"
  )


  include(FindPackageHandleStandardArgs)
  # handle the QUIETLY and REQUIRED arguments and set BSTools_FOUND to TRUE
  # if all listed variables are TRUE
  find_package_handle_standard_args(BSTools  DEFAULT_MSG
                                  BSTools_LIBRARIES BSTools_INCLUDE_DIR)
  if (BSTools_FOUND)
    message(STATUS "BSTools header files: ${BSTools_INCLUDE_DIR}")
    message(STATUS "BSTools library: ${BSTools_LIBRARIES}")
  endif (BSTools_FOUND)
  mark_as_advanced(BSTools_INCLUDE_DIRS BSTools_LIBRARIES)
endif (BSTools_LIBRARIES AND BSTools_INCLUDE_DIRS)

