#ifndef BSM_MESH_H
#define BSM_MESH_H

#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif
#include "pknum.h"
#include "bsmesh.h"

typedef struct {
 int spdimen; /** dimension of embedding space */
 int nv; /** number ov vertices */
 int nhe;  /** number of halfedges */
 int nfac; /** number of facets */
 BSMvertex *mv; /** array of vertices */
 double *vc; /** array of vertex positions */
 int *mvhei; /** array of indices of the halfedges having origins at 
                 the vertices */
 BSMhalfedge *mhe; /** array of halfedges */
 BSMfacet *mfac; /** array of facets */
 int *mfhei; /** array of indices of the halfedges forming the facets */
} BSMesh;

/** Allocates new, empty BSMesh structure */
BSMesh * bsm_NewMesh(void);

#define bsm_DeleteMesh(mesh) \
do { \
  bsm_ClearMesh(mesh);\
  PKV_FREE(mesh);\
} while (0);

void bsm_ClearMesh(BSMesh *mesh);

boolean bsm_ReadMesh(BSMesh *mesh, const char *filename);
boolean bsm_ReadMeshFromStream(BSMesh *mesh, FILE *stream);

boolean bsm_IsMeshEmpty(const BSMesh *mesh);
void bsm_WriteMeshInfo(const BSMesh *mesh, FILE *stream);

#ifdef __cplusplus
} /* extern C */
#endif

#endif /* ifndef BSM_MEHS_H  */
