
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <tiffio.h>

#include "pkvaria.h"
#include "pkgeom.h"
#include "pknum.h"
#include "xgedit.h"

#include "frames.h"

int           frame_cnt;
XImage        *image;
unsigned char *frame;

void InitXImage ( void )
{
  image = xge_CreateXImage ( xge_MAX_WIDTH, xge_MAX_HEIGHT );
  frame = (unsigned char*)malloc ( xge_MAX_HEIGHT*xge_MAX_WIDTH*3 );
  if ( !image || !frame )
    exit ( 1 );
} /*InitXImage*/

void InitFrameSequence ( void )
{
  frame_cnt = 0;
} /*InitFrameSequence*/

boolean WriteFrame ( short w, short h, short x, short y,   
                     int n, int k, boolean dual, boolean halfdual )
{
  char   dd;
  char   fn[80];
  TIFF   *tif;

  xge_GetImage ( w, h, x, y, frame, NULL );
  if ( dual ) dd = 'd';
  else if ( halfdual ) dd = 'h';
  else dd = 's';
  sprintf ( fn, "%c%d-%d-%04d.tif", dd, n, k, frame_cnt );
  printf ( "%s\n", fn );
  tif = TIFFOpen ( fn, "w" );
  if ( !tif )
    exit ( 1 );
  TIFFSetField ( tif, TIFFTAG_IMAGEWIDTH, w );
  TIFFSetField ( tif, TIFFTAG_IMAGELENGTH, h );
  TIFFSetField ( tif, TIFFTAG_BITSPERSAMPLE, 8 );
  TIFFSetField ( tif, TIFFTAG_SAMPLESPERPIXEL, 3 );
  TIFFSetField ( tif, TIFFTAG_ROWSPERSTRIP, h );
  TIFFSetField ( tif, TIFFTAG_COMPRESSION, COMPRESSION_LZW );
  TIFFSetField ( tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB );
  TIFFSetField ( tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG );
  TIFFSetField ( tif, TIFFTAG_XRESOLUTION, 150.0 );
  TIFFSetField ( tif, TIFFTAG_YRESOLUTION, 150.0 );
  TIFFSetField ( tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH );
  TIFFWriteEncodedStrip ( tif, 0, frame, 3*w*h );
  TIFFClose ( tif );
  frame_cnt ++;
  return frame_cnt < MAX_FRAMES;
} /*WriteFrame*/

