
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */   
/* //////////////////////////////////////////////////// */

static const char sh_version400[] =
  "#version 400\n";

static const char sh_localdef[] =
  "#define VERTEX_POS     0\n"
  "#define VERTEX_COLOR   1\n"
  "#define VERTEX_MARKING 2\n";

static const char sh_trans[] =
  "uniform ModelTransBlock {"
    "mat4 ModelMatrix;"
    "mat4 InvTModelMatrix;"
  "} transform;"

  "uniform ProjectionBlock {"
    "mat4 ViewMatrix;"
    "mat4 ProjectionMatrix;"
    "mat4 VPMatrix;"
    "vec4 EyePos;"
  "} project;";

static const char sh_v0[] =
  "layout(location=VERTEX_POS) in vec4 in_Position;"
  "layout(location=VERTEX_COLOR) in vec4 in_Color;"

  "out Vertex {"
    "vec4 Color;"
  "} Out;"

  "void main ( void )"
  "{"
    "gl_Position = in_Position;"
    "Out.Color = in_Color;"
  "}"; /*main*/

static const char sh_g0_0[] =
  "layout(triangles) in;"
  "layout(triangle_strip,max_vertices=3) out;"

  "in Vertex {"
    "vec4 Color;"
  "} In[];"

  "out NVertex {"
    "vec4 Color;"
    "vec3 Normal;"
    "vec3 Position;"
  "} Out;"

  "void main ( void )"
  "{"
    "int  i;"
    "vec3 v1, v2;"
    "vec4 Pos, Normal;"

    "v1 = gl_in[1].gl_Position.xyz/gl_in[1].gl_Position.w -"
         "gl_in[0].gl_Position.xyz/gl_in[0].gl_Position.w;"
    "v2 = gl_in[2].gl_Position.xyz/gl_in[2].gl_Position.w -"
         "gl_in[0].gl_Position.xyz/gl_in[0].gl_Position.w;";

static const char sh_g0_1[] =
    "Normal.xyz = cross ( v1, v2 );"
    "Normal.w = 0.0;"
    "Pos = transform.InvTModelMatrix * Normal;"
    "v1 = normalize ( Pos.xyz );"
    "for ( i = 0; i < 3; i++ ) {"
      "Pos = transform.ModelMatrix * gl_in[i].gl_Position;"
      "gl_Position = project.VPMatrix * Pos;"
      "Out.Position = Pos.xyz / Pos.w;"
      "Out.Normal = v1;"
      "Out.Color = In[i].Color;"
      "EmitVertex ();"
    "}"
    "EndPrimitive ();"
  "}"; /*main*/

static const char sh_v1[] =
  "layout(location=VERTEX_POS) in vec4 in_Position;"
  "layout(location=VERTEX_MARKING) in int in_Marking;"

  "out NVertex {"
    "vec4 Color;"
    "vec3 Normal;"
    "vec3 Position;"
  "} Out;"

  "void main ( void )"
  "{"
    "vec4 Pos;"

    "Pos = transform.ModelMatrix * in_Position;"
    "gl_Position = project.VPMatrix * Pos;"
    "if ( in_Marking != 0 )"
      "Out.Color = vec4 ( 1.0, 0.3, 0.0, 1.0 );"
    "else "
      "Out.Color = vec4 ( 1.0, 1.0, 0.0, 1.0 );"
    "Out.Normal = vec3 ( 0.0, 0.0, 0.0 );"
    "Out.Position = Pos.xyz / Pos.w;"
  "}"; /*main*/

static const char sh_f[] =
  "in NVertex {"
    "vec4 Color;"
    "vec3 Normal;"
    "vec3 Position;"
  "} In;"

  "out vec4 out_Color;"

  "void main ( void )"
  "{"
    "out_Color = In.Color;"
  "}"; /*main*/

static const char *_shader_v0[] =
  { sh_version400, sh_localdef, sh_v0 };

static const char *_shader_g0[] =
  { sh_version400, sh_trans, sh_g0_0, sh_g0_1 };

static const char *_shader_v1[] =
  { sh_version400, sh_localdef, sh_trans, sh_v1 };

static const char *_shader_f[] =
  { sh_version400, sh_f };

static const GLchar *UniformTBNames[3] =
  { "ModelTransBlock",
    "ModelTransBlock.ModelMatrix",
    "ModelTransBlock.InvTModelMatrix" };

static const GLchar *UniformPBNames[5] =
  { "ProjectionBlock",
    "ProjectionBlock.ViewMatrix",
    "ProjectionBlock.ProjectionMatrix",
    "ProjectionBlock.VPMatrix",
    "ProjectionBlock.EyePos" };

