#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/times.h>

#include "pkvaria.h"

#include "suHD.h"

boolean star_hints = false;

int NMaskBits ( unsigned short mask )
{
  int            i, j;
  unsigned short m;

  for ( i = j = 0, m = 0x01;  i < 16;  i++, m <<= 1 )
    if ( mask & m )
      j ++;
  return j;
} /*NMaskBits*/

boolean NumFromMask ( unsigned short mask, char *num )
{
  int            i, j;
  unsigned short m;

  for ( i = j = 0, m = 0x01;  i < 16;  i++, m <<= 1 )
    if ( mask & m ) { *num = i;  j++; }
  return j == 1;
} /*NumFromMask*/

char InitPrompts ( char hints[16][16], unsigned short prompts[16][16] )
{
  int            i, j, k, l;
  unsigned short mask;

  memset ( prompts, 0xFF, 256*sizeof(unsigned short) );
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ ) {
      if ( hints[i][j] >= 0 )
        prompts[i][j] = 0x0001 << hints[i][j];
      else {
        for ( k = 0; k < 16; k++ )
          if ( hints[i][k] >= 0) {
            mask = 0x0001 << hints[i][k];
            prompts[i][j] &= ~mask;
          }
        for ( k = 0; k < 16; k++ )
          if ( hints[k][j] >= 0 ) {
            mask = 0x0001 << hints[k][j];
            prompts[i][j] &= ~mask;
          }
        for ( k = i & 0xFC; k < (i & 0xFC)+4; k++ )
          for ( l = j & 0xFC; l < (j & 0xFC)+4; l++ )
            if ( hints[k][l] >= 0 ) {
              mask = 0x0001 << hints[k][l];
              prompts[i][j] &= ~mask;
            }
      }
    }
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( NMaskBits ( prompts[i][j] ) == 1 )
        return true;
  return false;
} /*InitPrompts*/

void UpdatePrompts0 ( char n, unsigned short prompts[16][16], int i, int j )
{
  int            k, l;
  unsigned short mask;

  prompts[i][j] = 0;
  mask = ~(0x0001 << n );
  for ( k = 0; k < 16; k++ ) {
    prompts[i][k] &= mask;
    prompts[k][j] &= mask;
  }
  for ( k = i & 0xFC; k < (i &0xFC)+4; k++ )
    for ( l = j & 0xFC; l < (j & 0xFC)+4; l++ )
      prompts[k][l] &= mask;
  prompts[i][j] = ~mask;
} /*UpdatePrompts0*/

boolean UpdatePrompts1 ( char su[16][16], unsigned short prompts[16][16] )
{
  boolean progress;
  int     i, j;
  char    num;

  progress = false;
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( su[i][j] < 0 )
        if ( NumFromMask ( prompts[i][j], &num ) ) {
          UpdatePrompts0 ( num, prompts, i, j );
          su[i][j] = num;
          progress = true;
        }

/*if ( progress ) printf ( "*" );*/

  return progress;
} /*UpdatePrompts1*/

boolean UpdateMustRow ( char su[16][16], unsigned short prompts[16][16],
                        int si, int sj, int sr )
{
  boolean        progress;
  int            i, j, k, l;
  unsigned short mask, newm;

  progress = false;
  mask = 0xFFFF;
  for ( i = 0, k = 4*si;  i < 4;  i++, k++ )
    if ( i != sr )
      for ( j = 0, l = 4*sj;  j < 4;  j++, l++ )
        if ( su[k][l] >= 0 )
          mask &= ~(0x0001 << su[k][l]);
  k = 4*si + sr;
  for ( i = 0; i < 4; i++ )
    if ( i != sj )
      for ( j = 0, l = 4*i;  j < 4;  j++, l++ )
        if ( su[k][l] >= 0 )
          mask &= ~(0x0001 << su[k][l]);
  if ( NMaskBits ( mask ) <= 4 ) {
    for ( i = 0, k = 4*si;  i < 4;  i++, k++ )
      if ( i != sr )
        for ( j = 0, l = 4*sj;  j < 4;  j++, l++ ) {
          newm = prompts[k][l] & ~mask;
          if ( newm != prompts[k][l] ) {
            prompts[k][l] = newm;
            progress = true;
          }
        }
    k = 4*si + sr;
    for ( i = 0; i < 4; i++ )
      if ( i != sj )
        for ( j = 0, l = 4*i;  j < 4;  j++, l++ ) {
          newm = prompts[k][l] & ~mask;
          if ( newm != prompts[k][l] ) {
            prompts[k][l] = newm;
            progress = true;
          }
        }
  }
  return progress;
} /*UpdateMustRow*/

boolean UpdateMustCol ( char su[16][16], unsigned short prompts[16][16],
                        int si, int sj, int sc )
{
  boolean        progress;
  int            i, j, k, l;
  unsigned short mask, newm;

  progress = false;
  mask = 0xFFFF;
  for ( j = 0, l = 4*sj;  j < 4;  j++, l++ )
    if ( j != sc )
      for ( i = 0, k = 4*si;  i < 4;  i++, k++ )
        if ( su[k][l] >= 0 )
          mask &= ~(0x0001 << su[k][l]);
  l = 4*sj + sc;
  for ( i = 0; i < 4; i++ )
    if ( i != si )
      for ( j = 0, k = 4*i;  j < 4;  j++, k++ )
        if ( su[k][l] >= 0 )
          mask &= ~(0x0001 << su[k][l]);
  if ( NMaskBits ( mask ) <= 4 ) {
    for ( j = 0, l = 4*sj;  j < 4;  j++, l++ )
      if ( j != sc )
        for ( i = 0, k = 4*si;  i < 4;  i++, k++ ) {
          newm = prompts[k][l] & ~mask;
          if ( newm != prompts[k][l] ) {
            prompts[k][l] = newm;
            progress = true;
          }
        }
    l = 4*sj + sc;
    for ( i = 0; i < 4; i++ )
      if ( i != si )
        for ( j = 0, k = 4*i;  j < 4;  j++, k++ ) {
          newm = prompts[k][l] & ~mask;
          if ( newm != prompts[k][l] ) {
            prompts[k][l] = newm;
            progress = true;
          }
        }
  }
  return progress;
} /*UpdateMustCol*/

boolean UpdatePrompts2 ( char su[16][16], unsigned short prompts[16][16] )
{
  boolean progress;
  int     i, j, k;

  progress = false;
  for ( i = 0; i < 4; i++ )
    for ( j = 0; j < 4; j++ )
      for ( k = 0; k < 4; k++ ) {
        progress |= UpdateMustRow ( su, prompts, i, j, k );
        if ( star_hints )
          progress |= UpdateMustCol ( su, prompts, i, j, k );
      }

/*if ( progress ) printf ( "." );*/

  return progress;
} /*UpdatePrompts2*/

boolean TestHints ( char su[16][16], char sustatus[16][16],
                    unsigned short prompts[16][16], char tsol[16][16],
                    char *unique )
{
  char    suh[16][16];
  boolean progress, pr;
  int     niter;
  int     i, j;

  memset ( suh, -1, 256*sizeof(char) );
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( sustatus[i][j] == SU_HINT )
        suh[i][j] = su[i][j];
  InitPrompts ( suh, prompts );
  memcpy ( tsol, suh, 256*sizeof(char) );
  pr = false;
  for ( niter = 0; niter < 16; niter ++ ) {
    progress = UpdatePrompts1 ( tsol, prompts );
    progress |= UpdatePrompts2 ( tsol, prompts );
    if ( !progress )
      break;
    else
      pr = true;
  }
/*  if ( pr ) printf ( "\n" );*/
  for ( i = 0; i < 16; i++ )
    for ( j = 0; j < 16; j++ )
      if ( NMaskBits ( prompts[i][j] ) > 1 ) {
        *unique = false;
        return pr;
      }
  *unique = true;
  return pr;
} /*TestHints*/

/* ////////////////////////////////////////////////////////////////////////// */
int FindHints ( char su[16][16], unsigned char hintpermut[256],
                unsigned char hpinv[256], char sustatus[16][16] )
{
  int            a, b, c;
  int            i, j;
  char           unique;
  unsigned short prompts[16][16];
  char           tsol[16][16];

  a = 0;  b = 256;
  while ( b-a > 1 ) {
    c = (b+a)/2;
    memset ( sustatus, SU_NONE, 256*sizeof(char) );
    for ( i = 0; i < c; i++ ) {
      j = hintpermut[i];
      sustatus[j/16][j%16] = SU_HINT;
    }
    TestHints ( su, sustatus, prompts, tsol, &unique );
    if ( unique )
      b = c;
    else
      a = c;
  }
  for ( i = 0; i < b; ) {
    j = hintpermut[i];
    sustatus[j/16][j%16] = SU_NONE;
    TestHints ( su, sustatus, prompts, tsol, &unique );
    if ( unique )
      { hintpermut[i] = hintpermut[--b];  hintpermut[b] = j; }
    else
      { sustatus[j/16][j%16] = SU_HINT;  i ++;  }
  }
  for ( i = 0; i < 256; i++ )
    hpinv[hintpermut[i]] = i;
  return b;
} /*FindHints*/

void TossUpHints ( char su[16][16], char sustatus[16][16],
                   unsigned char hintpermut[256], unsigned char hpinv[256],
                   int *nhints )
{
  struct tms    buffer;
  unsigned int  tmp, rnd;
  int           i, sw;
  unsigned char c;

  tmp = (unsigned int)times ( &buffer ) | 0x01;
  for ( i = 0; i < 10; i++ )
    tmp *= 16807;
  for ( i = 0; i < 256; i++ )
    hintpermut[i] = i;
  for ( i = 0; i < 255; i++ ) {
    tmp *= 16807;
    rnd = tmp ^ (tmp >> 4) ^ (tmp << 15);
    sw = i + rnd % (256-i);
    c = hintpermut[i];  hintpermut[i] = hintpermut[sw];  hintpermut[sw] = c;
  }
  for ( i = 0; i < 256; i++ )
    hpinv[hintpermut[i]] = i;
  *nhints = FindHints ( su, hintpermut, hpinv, sustatus );
} /*TossUpHints*/

/* ////////////////////////////////////////////////////////////////////////// */
boolean ExtraPrompts ( char usol[16][16], char sustatus[16][16] )
{
  unsigned short prompts[16][16];
  char           tsol[16][16];
  int            i, j;
  char           found, num;

  found = false;
  memcpy ( tsol, usol, 256*sizeof(char) );
  if ( InitPrompts ( tsol, prompts ) )
    goto find_eprompts;
  else if ( UpdatePrompts1 ( tsol, prompts ) )
    goto find_eprompts;
  else if ( UpdatePrompts2 ( tsol, prompts ) ) {
find_eprompts:
    for ( i = 0; i < 16; i++ )
      for ( j = 0; j < 16; j++ )
        if ( sustatus[i][j] == SU_NONE && NumFromMask ( prompts[i][j], &num ) ) {
          sustatus[i][j] = SU_EPROMPT;
          found = true;
        }
  }
  return found;
} /*ExtraPrompts*/

/* ////////////////////////////////////////////////////////////////////////// */
const char t1[] =
  "\\documentclass[12pt]{article}\n"
  "\\usepackage{times}\n"
  "\\pagestyle{empty}\n"
  "\\begin{document}\n";
const char t2[] =
  "\\centerline{%\n"
  "\\newcommand{\\bx}[1]{\\makebox[0pt][c]{#1}}\n"
  "\\begin{picture}(256,256)\n"
  "  \\multiput(0,16)(0,16){15}{\\line(1,0){256}}\n"
  "  \\multiput(16,0)(16,0){15}{\\line(0,1){256}}\n"
  "  \\thicklines\n"
  "  \\multiput(-0.4,0)(0,64){5}{\\line(1,0){256.7}}\n"
  "  \\multiput(0,0)(64,0){5}{\\line(0,1){256}}\n";
const char t3[] =
  "\\end{picture}}\n"
  "\\end{document}\n";

boolean ExportSu ( char *fn, char su[16][16], char sustatus[16][16] )
{
  FILE *f;
  int  i, j;

  f = fopen ( fn, "w" );
  if ( f ) {
    fprintf ( f, "%s", t1 );
    for ( i = 0; i < 16; i++ ) {
      fprintf ( f, "%c", '%' );
      for ( j = 0; j < 16; j++ )
        fprintf ( f, " %c", hd[(int)su[i][j]] );
      fprintf ( f, "\n" );
    }
    fprintf ( f, "%s", t2 );
    for ( i = 0; i < 16; i++ )
      for ( j = 0; j < 16; j++ )
        if ( sustatus[i][j] == SU_HINT )
          fprintf ( f, "  \\put(%d,%d){\\bx{%c}}\n",
                       8+16*j, 4+16*(15-i), hd[(int)su[i][j]] );
    fprintf ( f, "%s", t3 );
    fclose ( f );
    return true;
  }
  else
    return false;
} /*ExportSu*/

