
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

#define INBUFSIZE    1024
#define NAMEBUFSIZE  (BSF_MAX_NAME_LENGTH+1)

/* ///////////////////////////////////////////////////////////////////////// */
const char *bsf_keyword[BSF_NKEYWORDS] = /* this table must be sorted */
  { "BCurve",           /* alphabetically, uppercase before lowercase */
    "BPatch",           /* and the contents must match the numbers in */
    "BSCurve",          /* bsfprivate.h */
    "BSHole",
    "BSMesh",
    "BSPatch",
    "Euler_angles",
    "KLinkage",
    "art_param",
    "art_param_range",
    "btrans",
    "camera",
    "closed",
    "color",
    "colour",
    "cpoints",
    "cpointsmk",
    "degree",
    "depth",
    "dim",
    "domain",
    "etrans",
    "facetmk",
    "facets",
    "filename",
    "frame",
    "ftrans",
    "halfedgemk",
    "halfedges",
    "ident",
    "joint",
    "knots",
    "knots_u",
    "knots_v",
    "link",
    "name",
    "num_art_par",
    "num_joints",
    "num_links",
    "num_obj_ref",
    "num_objects",
    "num_vertices",
    "object",
    "object_ref",
    "parallel",
    "perspective",
    "points",
    "polyline",
    "position",
    "rational",
    "root",
    "rotation_v",
    "rotation_x",
    "rotation_y",
    "rotation_z",
    "scaling_x",
    "scaling_xyz",
    "scaling_y",
    "scaling_z",
    "sides",
    "spherical_product",
    "transform",
    "translation_v",
    "translation_x",
    "translation_y",
    "translation_z",
    "trimmed",
    "uniform",
    "vertices" };

/* ///////////////////////////////////////////////////////////////////////// */
boolean bsf_AppIdentFind ( pks_scanner *sc )
{
  int r;

  if ( (r = pks_BinSearchNameTable ( BSF_NKEYWORDS, 0,
                       bsf_keyword, sc->nextname ) ) != PKS_SYMB_ERROR ) {
    sc->id_space = ID_SPACE_BSFILE;
    sc->nextinteger = r;
  }
  else if ( (r = pks_BinSearchNameTable ( pkscalc_NUMNAMES, 0, pkscalc_fnames,
                       sc->nextname ) ) != PKS_SYMB_ERROR ) {
    sc->id_space = ID_SPACE_CALC;
    sc->nextinteger = r;
  }
  else {
    sc->id_space = ID_SPACE_NONE;
    sc->nextinteger = -1;
  
  }
  return true;
} /*bsf_AppIdentFind*/

boolean _bsf_OpenInputFile ( const char *filename, bsf_Reader *reader )
{
  if ( pks_InitPreprocessor ( &reader->sc, PKS_OPT_C_OPERATORS,
               0, BSF_MAX_NAME_LENGTH, "%", "\n", NULL,
               NULL, bsf_AppIdentFind, NULL, (void*)reader ) ) {
    if ( pks_OpenInputFile ( &reader->sc, filename ) ) {
      reader->errcode = 0;
      reader->sc.GetNextSymbol ( &reader->sc );
      return true;
    }
    else
      pks_ShutDownPreprocessor ( &reader->sc );
  }
  return false;
} /*_bsf_OpenInputFile*/

void bsf_PrintErrorLocation ( bsf_Reader *reader )
{
  fprintf ( stderr, "Error detected at line %d, column %d\n",
            reader->sc.nsline+1, reader->sc.nscolumn+1 );
} /*bsf_PrintErrorLocation*/

