
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadIntNumber ( pks_scanner *sc, int *number )
{
  switch ( sc->nextsymbol ) {
case PKS_SYMB_MINUS:
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol != PKS_SYMB_INTEGER )
      return false;
    else {
      *number = -sc->nextinteger;
      sc->GetNextSymbol ( sc );
      return true;
    }

case PKS_SYMB_PLUS:
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol != PKS_SYMB_INTEGER )
      return false;
    else {
      *number = sc->nextinteger;
      sc->GetNextSymbol ( sc );
      return true;
    }

case PKS_SYMB_INTEGER:
    *number = sc->nextinteger;
    sc->GetNextSymbol ( sc );
    return true;

default:
    return false;
  }
} /*bsf_ReadIntNumber*/

boolean bsf_ReadDoubleNumber ( pks_scanner *sc, double *number )
{
  switch ( sc->nextsymbol ) {
case PKS_SYMB_MINUS:
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol == PKS_SYMB_INTEGER || sc->nextsymbol == PKS_SYMB_FLOAT ) {
      *number = -sc->nextfloat;
      sc->GetNextSymbol ( sc );
      return true;
    }
    else
      return false;

case PKS_SYMB_PLUS:
    sc->GetNextSymbol ( sc );
    if ( sc->nextsymbol != PKS_SYMB_INTEGER && sc->nextsymbol != PKS_SYMB_FLOAT )
      return false;
case PKS_SYMB_INTEGER:
case PKS_SYMB_FLOAT:
    *number = sc->nextfloat;
    sc->GetNextSymbol ( sc );
    return true;

default:
    return false;
  }
} /*bsf_ReadDoubleNumber*/

boolean bsf_ReadIdent ( pks_scanner *sc, int *ident )
{
  if ( sc->nextsymbol != PKS_SYMB_IDENT )
    return false;
  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_INTEGER )
    return false;
  if ( ident )
    *ident = sc->nextinteger;
  sc->GetNextSymbol ( sc );
  return true;
} /*bsf_ReadIdent*/

