
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

void bsf_WritePointd ( bsf_Writer *writer, int spdimen, const double *point )
{
  double defaultp[4] = {0.0,0.0,0.0,1.0};
  int i, k;

        /* do not write out default coordinates - y == 0, z == 0, w == 1 */
  k = spdimen;
  if ( k <= 4 ) {
    while ( k > 1 && point[k-1] == defaultp[k-1] )
      k--;
  }
  writer->current_length += fprintf ( writer->output, "{" );
  for ( i = 0; i < k-1; i++ ) {
    bsf_WriteDoubleNumber ( writer, point[i] );
    writer->current_length += fprintf ( writer->output, "," );
  }
  bsf_WriteDoubleNumber ( writer, point[k-1] );
  writer->current_length += fprintf ( writer->output, "}" );
} /*bsf_WritePointd*/

void bsf_WritePointsd ( bsf_Writer *writer,
                        int spdimen, int cols, int rows, int pitch,
                        const double *points )
{
  int i, j, k, l;
  int sci;

  sci = writer->current_indentation;
  BSFwci
  writer->current_length += fprintf ( writer->output, "{" );
  BSFeol
/*  writer->current_indentation += 2;*/
  BSFwci
  for ( i = k = 0;  i < cols;  i++, k += pitch ) {
    for ( j = 0, l = k;  j < rows;  j++, l += spdimen ) {
      bsf_WritePointd ( writer, spdimen, &points[l] );
      if ( j < rows-1 ) {
        writer->current_length += fprintf ( writer->output, "," );
        BSFdol
        BSFwci
      }
    }
    if ( i < cols-1 ) {
      writer->current_length += fprintf ( writer->output, "," );
      BSFeol
      writer->current_indentation -= 2;
      BSFwci
      writer->current_length += fprintf ( writer->output, "%c %d:", '%', i+1 );
      writer->current_indentation += 2;
      BSFeol
      BSFwci
    }
    else {
      writer->current_length += fprintf ( writer->output, "}" );
      BSFeol
    }
  }
  writer->current_indentation = sci;
} /*bsf_WritePointsd*/

void bsf_WriteTrans3d ( bsf_Writer *writer, const trans3d *tr )
{
  int i, j;
  writer->current_length += fprintf ( writer->output, "{" );
  for ( i = 0; i < 3; i++ )
    for ( j = 0; j < 4; j++ ) {
      bsf_WriteDoubleNumber ( writer, tr->U1.a[i][j] );
      if ( i < 2 || j < 3 ) {
        writer->current_length += fprintf ( writer->output, "," );
        BSFdol
      }
    }
  writer->current_length += fprintf ( writer->output, "}" );
} /*bsf_WriteTrans3d*/

static void _bsf_WriteMK ( bsf_Writer *writer,
                           int nmk, const unsigned int *mk, int key )
{
  int i;
  int sci;

  if ( mk && nmk > 0 ) {
    sci = writer->current_indentation;
    BSFwci
    writer->current_indentation += 2;
    writer->current_length +=
          fprintf ( writer->output, "%s {",
                    bsf_keyword[key] );
    for ( i = 0; i < nmk-1; i++ ) {
      writer->current_length += fprintf ( writer->output, "%d,", mk[i] );
      BSFdol
      BSFwci
    }
    writer->current_length += fprintf ( writer->output, "%d}", mk[nmk-1] );
    BSFeol
    writer->current_indentation = sci;
  }
} /*_bsf_WriteMK*/

void bsf_WritePointsMK ( bsf_Writer *writer, int npoints, const unsigned int *mk )
{
  _bsf_WriteMK ( writer, npoints, mk, BSF_KEY_CPOINTSMK );
} /*bsf_WritePointsMK*/

void bsf_WriteHEdgesMK ( bsf_Writer *writer, int nhe, const unsigned int *hemk )
{
  _bsf_WriteMK ( writer, nhe, hemk, BSF_KEY_HALFEDGEMK );
} /*bsf_WriteHEdgesMK*/

void bsf_WriteFacetsMK ( bsf_Writer *writer, int nfac, const unsigned int *fmk )
{
  _bsf_WriteMK ( writer, nfac, fmk, BSF_KEY_FACETMK );
} /*bsf_WriteFacetsMK*/

