
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

static boolean IsUniform ( int lastknot, const double *knots )
{
  int i;

  for ( i = 0; i <= lastknot; i++ )
    if ( knots[i] != (double)i )
      return false;
  return true;
} /*IsUniform*/

void bsf_WriteKnotSequenced ( bsf_Writer *writer,
                              int lastknot, const double *knots, boolean closed )
{
  int i;
  int sci;

  if ( lastknot < 1 )
    return;

  sci = writer->current_indentation;
  writer->current_indentation += 2;
  if ( closed ) {
    writer->current_length +=
          fprintf ( writer->output, " %s",
                    bsf_keyword[BSF_KEY_CLOSED] );
  }

  if ( IsUniform ( lastknot, knots ) ) {
    writer->current_length +=
          fprintf ( writer->output, " %s %d",
                    bsf_keyword[BSF_KEY_UNIFORM], lastknot );
  }
  else {
    writer->current_length += fprintf ( writer->output, " {" );
    for ( i = 0; i < lastknot; i++ ) {
      bsf_WriteDoubleNumber ( writer, knots[i] );
      writer->current_length += fprintf ( writer->output, "," );
      BSFdol
      BSFwci
    }
    bsf_WriteDoubleNumber ( writer, knots[lastknot] );
    writer->current_length += fprintf ( writer->output, "}" );
  }
  BSFeol
  writer->current_indentation = sci;
} /*WriteKnotSequence*/

