
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2010, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_WriteBSMeshd ( bsf_Writer *writer,
                           int spdimen, int cpdimen, boolean rational, int degree,
                           int nv, const BSMvertex *mv, const int *mvhei,
                           const double *vc,
                           int nhe, const BSMhalfedge *mhe,
                           int nfac, const BSMfacet *mfac, const int *mfhei,
                           const char *name, int ident,
                           bsf_WriteAttr_fptr WriteAttr, void *userData )
{
  int sci;
  int i, j, d, l;

  sci = writer->current_indentation;
  writer->current_length += fprintf ( writer->output, "%s\n", "%B-spline mesh" );
  writer->current_length += fprintf ( writer->output, "%s {",
          bsf_keyword[BSF_KEY_BSMESH] );
  BSFeol
  writer->current_indentation += 2;
  if ( name && *name ) {
    BSFwci
    writer->current_length += fprintf ( writer->output, "%s \"%s\"",
              bsf_keyword[BSF_KEY_NAME], name );
    BSFeol
  }
  BSFwci
  bsf_WriteSpaceDim ( writer, spdimen, rational );
  if ( degree >= -1 )
    bsf_WriteCurveDegree ( writer, degree );
        /* write out the mesh vertices */
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d {",
            bsf_keyword[BSF_KEY_VERTICES],
            nv );
  BSFeol
  writer->current_indentation += 2;
  BSFwci
  for ( i = 0; i < nv; i++ ) {
    writer->current_length += fprintf ( writer->output, "{" );
    d = mv[i].degree;
    l = mv[i].firsthalfedge;
    for ( j = 0; j < d; j++ ) {
      writer->current_length += fprintf ( writer->output, "%d", mvhei[l+j] );
      if ( j < d-1 )
        writer->current_length += fprintf ( writer->output, ", " );
    }
    writer->current_length += fprintf ( writer->output, " " );
    bsf_WritePointd ( writer, cpdimen, &vc[i*spdimen] );
    writer->current_length += fprintf ( writer->output, "}" );
    if ( i < nv-1 ) {
      writer->current_length += fprintf ( writer->output, "," );
      BSFdol
      BSFwci
    }
  }
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  writer->current_indentation -= 2;
        /* write out the mesh halfedges */
  BSFwci
  writer->current_length += fprintf ( writer->output, "%s %d {",
            bsf_keyword[BSF_KEY_HALFEDGES],
            nhe );
  BSFeol
  writer->current_indentation += 2;
  BSFwci
  for ( i = 0; i < nhe; i++ ) {
    writer->current_length += fprintf ( writer->output, "{%d, %d, %d, %d}",
              mhe[i].v0, mhe[i].v1, mhe[i].facetnum, mhe[i].otherhalf );
    if ( i < nhe-1 ) {
      writer->current_length += fprintf ( writer->output, "," );
      BSFdol
      BSFwci
    }
  }
  writer->current_length += fprintf ( writer->output, "}");
  BSFeol
  writer->current_indentation -= 2;
  BSFwci
        /* write out the mesh facets */
  writer->current_length += fprintf ( writer->output, "%s %d {",
            bsf_keyword[BSF_KEY_FACETS],
            nfac );
  BSFeol
  writer->current_indentation += 2;
  BSFwci
  for ( i = 0; i < nfac; i++ ) {
    writer->current_length += fprintf ( writer->output, "{" );
    d = mfac[i].degree;
    l = mfac[i].firsthalfedge;
    for ( j = 0; j < d; j++ ) {
      writer->current_length += fprintf ( writer->output, "%d", mfhei[l+j] );
      if ( j < d-1 )
        writer->current_length += fprintf ( writer->output, ", " );
    }
    writer->current_length += fprintf ( writer->output, "}" );
    if ( i < nfac-1 ) {
      writer->current_length += fprintf ( writer->output, "," );
      BSFdol
      BSFwci
    }
  }
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  writer->current_indentation -= 2;
  BSFwci
  if ( WriteAttr )
    WriteAttr ( writer, userData );
  writer->current_indentation = sci;
  BSFwci
  writer->current_length += fprintf ( writer->output, "}" );
  BSFeol
  return true;
} /*bsf_WriteBSMeshd*/

