
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2014, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h> 
#include <math.h> 
#include <malloc.h>
#include <string.h> 
#include <ctype.h>  

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadPolyline4d ( bsf_Reader *reader,
                             int maxvert, int *nvert, point4d *vert,
                             int *spdimen, boolean *rational, boolean *closed,
                             char *name, int *ident )
{
  boolean _name, _dimen, _id, _vert, _closed;
  int     nv, dim, cpdimen;

  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
       reader->sc.id_space != ID_SPACE_BSFILE ||
       reader->sc.nextinteger != BSF_KEY_POLYLINE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );

        /* nothing has been read in yet */
  _name = _dimen = _id = _vert = _closed = false;
  nv = 0;
  if ( name )   *name = 0;
  if ( *ident ) *ident = -1;
  *rational = *closed = false;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_NAME:
        if ( _name )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_STRING ) {
          strcpy ( name, reader->sc.nextname );
          reader->sc.GetNextSymbol ( &reader->sc );
        }
        else
          goto failure;
        _name = true;
        break;

  case BSF_KEY_IDENT:
        if ( _id )
          goto failure;
        if ( !bsf_ReadIdent ( &reader->sc, ident ) )
          goto failure;
        _id = true;
        break;

  case BSF_KEY_CLOSED:
        if ( _closed )
          goto failure;
        _closed = true;
        *closed = true;
        reader->sc.GetNextSymbol ( &reader->sc );
        break;

  case BSF_KEY_DIM:
        if ( _dimen )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadSpaceDim ( &reader->sc, 4, &dim ) )
          goto failure;
        _dimen = true;
        break;

  case BSF_KEY_POINTS:
        if ( _vert )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        nv = bsf_ReadPointsd ( &reader->sc, 4, maxvert, &vert[0].x, &cpdimen );
        if ( !nv )
          goto failure;
        *nvert = nv;
        _vert = true;
        break;

  case BSF_KEY_RATIONAL:
        reader->sc.GetNextSymbol ( &reader->sc );
        *rational = true;
        break;

        /* optional attributes */
  case BSF_KEY_CPOINTSMK:
        if ( !_bsf_ReadCPMark ( reader, maxvert ) )
          goto failure;
        break;

  case BSF_KEY_COLOR:
  case BSF_KEY_COLOUR:
        if ( !_bsf_ReadColour ( reader ) )
          goto failure;
        break;

  default:
        goto failure;
      }
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      if ( !_dimen || !_vert )
        goto failure;
      if ( (*rational && dim < cpdimen-1) || (!*rational && dim < cpdimen) )
        goto failure;
      *spdimen = dim;
      return true;

default:
      goto failure;
    }
  }

failure:
  return false;
} /*bsf_ReadPolyline4d*/

