
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <string.h>
#include <math.h>

#include "pknum.h"
#include "pkgeom.h"

/* ////////////////////////////////////////////////////////////////////////// */
/* construct a rotation/translation, mapping point p0 -> q0,                  */
/* line p0p1 ->q0q1, and plane p0p1p2 ->q0q1q2; triples p0,p1,p2 and q0,q1,q2 */
/* must not be collinear                                                      */
boolean SetRotationToPos3f ( point3f *p0, point3f *p1, point3f *p2,
                             point3f *q0, point3f *q1, point3f *q2, trans3f *tr )
{
#define TOL 1.0e-15
  vector3f v1, v2, v3, w1, w2, w3;
  float    t, Q0[9], Q1[9];
  point3f  r;

  SubtractPoints3f ( p1, p0, &v1 );
  t = DotProduct3f ( &v1, &v1 );
  if ( t < TOL*TOL )
    return false;
  MultVector3f ( 1.0/sqrt(t), &v1, &v1 );
  SubtractPoints3f ( p2, p0, &v2 );
  t = DotProduct3f ( &v1, &v2 );
  AddVector3Mf ( &v2, &v1, -t, &v2 );
  t = DotProduct3f ( &v2, &v2 );
  if ( t < TOL*TOL )
    return false;
  MultVector3f ( 1.0/sqrt(t), &v2, &v2 );
  CrossProduct3f ( &v1, &v2, &v3 );
  Q0[0] = v1.x;  Q0[1] = v1.y;  Q0[2] = v1.z;
  Q0[3] = v2.x;  Q0[4] = v2.y;  Q0[5] = v2.z;
  Q0[6] = v3.x;  Q0[7] = v3.y;  Q0[8] = v3.z;

  SubtractPoints3f ( q1, q0, &w1 );
  t = DotProduct3f ( &w1, &w1 );
  if ( t < TOL*TOL )
    return false;
  MultVector3f ( 1.0/sqrt(t), &w1, &w1 );
  SubtractPoints3f ( q2, q0, &w2 );
  t = DotProduct3f ( &w1, &w2 );
  AddVector3Mf ( &w2, &w1, -t, &w2 );
  t = DotProduct3f ( &w2, &w2 );
  if ( t < TOL*TOL )
    return false;
  MultVector3f ( 1.0/sqrt(t), &w2, &w2 );
  CrossProduct3f ( &w1, &w2, &w3 );
  Q1[0] = w1.x;  Q1[1] = w2.x;  Q1[2] = w3.x;
  Q1[3] = w1.y;  Q1[4] = w2.y;  Q1[5] = w3.y;
  Q1[6] = w1.z;  Q1[7] = w2.z;  Q1[8] = w3.z;

  IdentTrans3f ( tr );
  pkn_MultMatrixf ( 3, 3, 3, Q1, 3, 3, Q0, 4, &tr->U0.a11 );
  TransPoint3f ( tr, p0, &r );
  tr->U0.a14 = q0->x - r.x;
  tr->U0.a24 = q0->y - r.y;
  tr->U0.a34 = q0->z - r.z;
  return true;
#undef TOL
} /*SetRotationToPos3f*/

