
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <string.h>
#include <math.h>

#include "pkvaria.h"

void pkv_RGBtoHSVd ( double rgb[3], double hsv[3] )
{
  double r, g, b, mx, mn, d;

  r = max ( 0.0, rgb[0] );  r = min ( 1.0, r );
  g = max ( 0.0, rgb[1] );  g = min ( 1.0, g );
  b = max ( 0.0, rgb[2] );  b = min ( 1.0, b );
  if ( r > g ) { mx = max ( r, b );  mn = min ( g, b ); }
          else { mx = max ( g, b );  mn = min ( r, b ); }
  d = mx - mn;
  hsv[2] = mx;
  hsv[1] = mx > 0.0 ? d/mx : 0.0;
  if ( hsv[1] > 0.0 ) {
    if ( rgb[0] == mx )      hsv[0] = (rgb[1] - rgb[2])/d;
    else if ( rgb[1] == mx ) hsv[0] = 2.0 + (rgb[2] - rgb[0])/d;
    else                     hsv[0] = 4.0 + (rgb[0] - rgb[1])/d;
    if ( hsv[0] < 0.0 ) hsv[0] += 6.0;
    hsv[0] *= PI/3.0;
  }
} /*pkv_RGBtoHSVd*/

void pkv_HSVtoRGBd ( double hsv[3], double rgb[3] )
{
  double h, s, v, a, b, c, f;
  int    i;

  h = hsv[0];
  s = max ( 0.0, hsv[1] );  s = min ( 1.0, s );
  v = max ( 0.0, hsv[2] );  v = min ( 1.0, v );
  if ( s > 0.0 ) {
    while ( h < 0.0 ) h += 2.0*PI;
    while ( h >= 2.0*PI ) h -= 2.0*PI;
    h *= 3.0/PI;  i = (int)h;  f = h - (double)i;
    a = v*(1-s);  b = v*(1.0-s*f);  c = v*(1.0-s*(1.0-f));
    switch ( i ) {
case 0:  rgb[0] = v;  rgb[1] = c;  rgb[2] = a;  break;
case 1:  rgb[0] = b;  rgb[1] = v;  rgb[2] = a;  break;
case 2:  rgb[0] = a;  rgb[1] = v;  rgb[2] = c;  break;
case 3:  rgb[0] = a;  rgb[1] = b;  rgb[2] = v;  break;
case 4:  rgb[0] = c;  rgb[1] = a;  rgb[2] = v;  break;
case 5:  rgb[0] = v;  rgb[1] = a;  rgb[2] = b;  break;
    }
  }
  else
    rgb[0] = rgb[1] = rgb[2] = v;
} /*pkv_HSVtoRGBd*/

