
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"

#include "xgedit.h"
#include "xgeprivate.h"

static boolean _xge_hsv_dial = false;

static void _SetPixel ( short w, byte *data,
                        short x, short y, byte r, byte g, byte b )
{
  int z;

  z = 3*((int)w*y+x);
  data[z]   = r;
  data[z+1] = g;
  data[z+2] = b;
} /*_SetPixel*/

static void _DrawDialImage ( short w, short h, byte *data,
                             xgecolour_int c )
{
  byte  r1, g1, b1, r2, g2, b2;
  short x, y;
  float X, Y, R;

  xge_GetPixelColour ( xgec_MENU_BACKGROUND, &r1, &g1, &b1 );
  xge_GetPixelColour ( c, &r2, &g2, &b2 );
  for ( y = 0; y < h; y++ ) {
    Y = (float)(y-(h-1)/2);
    for ( x = 0; x < w; x++ ) {
      X = (float)(x-(w-1)/2);
      R = (float)sqrt ( X*X+Y*Y );
      if ( R >= h/2.0-3.0 )
        _SetPixel ( w, data, x, y, r1, g1, b1 );
      else if ( R >= h/2.0-6 )
        _SetPixel ( w, data, x, y, 255, 255, 255 );
      else
        _SetPixel ( w, data, x, y, r2, g2, b2 );
    }
  }
} /*_DrawDialImage*/

boolean _xge_InitDialImages ( void )
{
  void  *sp;
  short imw, imh;
  byte  *data;

  if ( _xge_hsv_dial )
    return true;
  if ( !_xge_InitCanvas () )
    return false;
  sp = pkv_GetScratchMemTop ();
  imw = 3*XGE_DIALICON_W+4;
  imh = 3*XGE_DIALICON_H+4;
  data = pkv_GetScratchMem ( 3*imw*imh );
  if ( !data )
    return false;
  _DrawDialImage ( imw, imh, data, xgec_DIAL_A );
  _xge_DrawOnCanvasAA ( XGE_DIALICON_W, XGE_DIALICON_H,
                        XGE_DIALICON_XA, XGE_HSVDIALICON_YA, imw, imh, data );
  _DrawDialImage ( imw, imh, data, xgec_DIAL_B );
  _xge_DrawOnCanvasAA ( XGE_DIALICON_W, XGE_DIALICON_H,
                        XGE_DIALICON_XB, XGE_DIALICON_YB, imw, imh, data );
  _xge_hsv_dial = true;
  pkv_SetScratchMemTop ( sp );
  return true;
} /*_xge_InitDialImages*/

