
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"

#include "xgedit.h"
#include "xgeprivate.h"

/* ////////////////////////////////////////////////////////////////////////// */
boolean xge_FilenameCorrect ( char *filename )
{
            /* a proper procedure of verification of file names ought to be */
            /* written here in order to improve the software reliability */
            /* now I am checking only that the file name is non-empty */
  return filename[0] != 0;
} /*xge_FilenameCorrect*/

void xge_SetCurrentDir ( xge_filemenu *fm, boolean open )
{
  int l;

  getcwd ( fm->fullpath, XGE_MAX_PATH_LENGTH+1 );
  l = strlen ( fm->fullpath );
  if ( l <= XGE_SHORT_PATH_LENGTH )
    strcpy ( fm->shortpath, fm->fullpath );
  else {
    strcpy ( fm->shortpath, "..." );
    strcpy ( &fm->shortpath[3], &fm->fullpath[l-XGE_SHORT_PATH_LENGTH+3] );
  }
  if ( open ) {
    xge_SetupFileList ( &fm->filelist1, ".", fm->filter, fm->showhidden );
    xge_ListBoxSetCurrentItem ( &fm->filelist1, fm->lastfileitem1 );
    xge_SetupDirList ( &fm->dirlist1, ".", NULL, fm->showhidden, fm->prevpath );
  }
  else {
    xge_SetupFileList ( &fm->filelist2, ".", fm->filter, fm->showhidden );
    xge_ListBoxSetCurrentItem ( &fm->filelist2, fm->lastfileitem2 );
    xge_SetupDirList ( &fm->dirlist2, ".", NULL, fm->showhidden, fm->prevpath );
  }
} /*xge_SetCurrentDir*/

void xge_ChangeDir ( xge_filemenu *fm, boolean open )
{
  char        *newdirname;
  xge_listbox *dl;

  dl = open ? &fm->dirlist1 : &fm->dirlist2;
  newdirname = &dl->itemstr[dl->itemind[dl->currentitem]];
  strncpy ( fm->prevpath, fm->fullpath, XGE_MAX_PATH_LENGTH );
  fm->prevpath[XGE_MAX_PATH_LENGTH] = 0;
  if ( !chdir ( newdirname ) ) {
    fm->lastfileitem1 = fm->lastfileitem2 = 0;
    xge_SetCurrentDir ( fm, open );
  }
} /*xge_ChangeDir*/

void xge_ChangeDirAlt ( xge_filemenu *fm, short x, boolean open )
{
  short l1, l2;

  strncpy ( fm->prevpath, fm->fullpath, XGE_MAX_PATH_LENGTH );
  fm->prevpath[XGE_MAX_PATH_LENGTH] = 0;
  l1 = strlen ( fm->fullpath );
  l2 = strlen ( fm->shortpath );
  x += l2-l1;
  while ( x < l1 && fm->fullpath[x] != '/' )
    x ++;
  if ( x < l1 ) {
    fm->fullpath[x] = 0;
    if ( chdir ( fm->fullpath ) )
      fm->fullpath[x] = '/';
    else {
      fm->lastfileitem1 = fm->lastfileitem2 = 0;
      xge_SetCurrentDir ( fm, open );
    }
  }
} /*xge_ChangeDirAlt*/

