
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2012, 2016                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "xgedit.h"
#include "xgeprivate.h"

/* ///////////////////////////////////////////////////////////////////////// */
void xge_DrawQuatRotBalld ( xge_widget *er, boolean onscreen )
{
  xge_quatrotballd *qball;
  short            xc, yc, r2, x, y;
  char             *title;
  trans3f          tr;

  qball = (xge_quatrotballd*)er->data0;
  Trans3dTof ( qball->tr, &tr );
  _xge_DrawQuaternionBallIcon ( &tr, er->state != xgestate_NOTHING );
  XPutImage ( xgedisplay, xgepixmap, xgegc, _xge_canvas,
              XGE_QBALLICON_X, XGE_QBALLICON_Y, er->x, er->y,
              XGE_QBALLICON_W, XGE_QBALLICON_H );
  r2 = XGE_QBALLICON_W/2;
  xc = er->x+r2-1;
  yc = er->y+r2-1;
  if ( (title = er->data1) ) {
    xgeSetForeground ( xgec_WIDGET_FOREGROUND );
    if ( er->w > er->h )  /* title aside */
      { x = (short)(xc+r2+4);  y = (short)(yc+4); }
    else                  /* title below */
      { x = (short)(xc-6*strlen(title)/2+1 );  y = (short)(yc+r2+13); }
    xgeDrawString ( title, x, y );
  }
  if ( onscreen )
    xgeCopyRectOnScreen ( er->w, er->h, er->x, er->y );
} /*xge_DrawQuatRotBalld*/

static boolean xge_QuatRotBalldRotateIt ( xge_widget *er, short x, short y )
{
  xge_quatrotballd *qball;
  short            dx, dy, xc, yc;
  vector3d         a, b, v;
  double           d, e, phi, s, c;
  quaterniond      q;

  dx = x - xge_xx;
  dy = y - xge_yy;
  if ( !dx && !dy )
    return false;
  qball = (xge_quatrotballd*)er->data0;
  xc = er->x+XGE_QBALLICON_W/2-1;
  yc = er->y+XGE_QBALLICON_W/2-1;
  switch ( er->state ) {
case xgestate_QUATROT_TURNING1:
    SetVector3d ( &a, (double)(x-xc)/(double)XGE_QUBALL_R,
                      (double)(yc-y)/(double)XGE_QUBALL_R, 0.0 );
    break;
case xgestate_QUATROT_TURNING2:
    SetVector3d ( &a, 0.0, 0.0, 0.0 );
    break;
case xgestate_QUATROT_TURNING3:
    if ( !qball->axis_ok ) {
      SetVector3d ( &qball->axis, (double)(x-xc)/(double)XGE_QUBALL_R,
                        (double)(yc-y)/(double)XGE_QUBALL_R, 0.0 );
      qball->axis_ok = true;
    }
    a = qball->axis;
    break;
default:
    return false;
  }
  SetVector3d ( &b, (double)(x-xge_xx)/(double)XGE_QUBALL_R,
                    (double)(xge_yy-y)/(double)XGE_QUBALL_R, 0.0 );
  OrtVector3d ( &b, &a, &a );
  d = a.x*a.x + a.y*a.y;
  e = b.x*b.x + b.y*b.y;
  if ( d >= 1.0 ) {
    phi = atan2 ( sqrt(e), sqrt(d) );
    NormalizeVector3d ( &a );
    if ( a.x*b.y - a.y*b.x >= 0.0 )
      SetVector3d ( &v, 0.0, 0.0, 1.0 );
    else
      SetVector3d ( &v, 0.0, 0.0, -1.0 );
  }
  else {
    a.z = sqrt ( 1.0 - d );
    CrossProduct3d ( &a, &b, &v );
    NormalizeVector3d ( &v );
    phi = sqrt ( e );
  }
  c = cos ( 0.5*phi );
  s = sin ( 0.5*phi );
  SetQuaterniond ( &q, c, s*v.x, s*v.y, s*v.z );
  QuaternionMultd ( &q, qball->q, qball->q );
  IdentTrans3d ( qball->tr );
  QuaternionRotTrans3d ( qball->tr, qball->q );
  xge_xx = x;
  xge_yy = y;
  return true;
} /*xge_QuatRotBalldRotateIt*/

static boolean xge_QuatRotBalldRotateWheel ( xge_widget *er, int key )
{
#define ANG (PI/180.0)
  xge_quatrotballd *qball;
  double           c, s;
  quaterniond      q;

  qball = (xge_quatrotballd*)er->data0;
  c = cos ( 0.5*ANG );
  s = sin ( 0.5*ANG );
  switch ( er->state ) {
case xgestate_NOTHING:
    if ( key & xgemouse_WHEELFW_CHANGE )
      SetQuaterniond ( &q, c, -s, 0.0, 0.0 );
    else
      SetQuaterniond ( &q, c, s, 0.0, 0.0 );
    break;
case xgestate_QUATROT_TURNING1:
    if ( key & xgemouse_WHEELFW_CHANGE )
      SetQuaterniond ( &q, c, 0.0, s, 0.0 );
    else
      SetQuaterniond ( &q, c, 0.0, -s, 0.0 );
    break;
case xgestate_QUATROT_TURNING2:
    if ( key & xgemouse_WHEELFW_CHANGE )
      SetQuaterniond ( &q, c, 0.0, 0.0, s );
    else
      SetQuaterniond ( &q, c, 0.0, 0.0, -s );
    break;
default:
    return false;
  }
  if ( qball->insert )
    QuaternionMultd ( qball->q, &q, qball->q );
  else
    QuaternionMultd ( &q, qball->q, qball->q );
  IdentTrans3d ( qball->tr );
  QuaternionRotTrans3d ( qball->tr, qball->q );
  return true;
#undef ANG
} /*xge_QuatRotBalldRotateWheel*/

static void _xge_QuatRotBalldDrawSpecial ( xge_widget *er, boolean onscreen )
{
  xge_quatrotballd *qball;
  short            xc, yc;
  trans3f          tr;

  qball = (xge_quatrotballd*)er->data0;
  xc = qball->er->x+XGE_QBALLICON_W/2-1;
  yc = qball->er->y+XGE_QBALLICON_W/2-1;
  xgeSetForeground ( xgec_QUATROTBALL_F );
  xgeDrawArc ( 2*XGE_QUBALL_R+1, 2*XGE_QUBALL_R+1,
               xc-XGE_QUBALL_R, yc-XGE_QUBALL_R, 0, 360*64 );
  xgeSetForeground ( xgec_QUATROTBALL_D );
  Trans3dTof ( qball->tr, &tr );
  _xge_QuatRotBallDrawCircles ( xc, yc, XGE_QUBALL_R, &tr,
                                xge_OutPixels, NULL );
  if ( onscreen )
    xgeCopyRectOnScreen ( er->w, er->h, er->x, er->y );
} /*_xge_QuatRotBalldDrawSpecial*/

#ifdef USE_XEXT_SHAPE
static void _xge_QuatRotBalldMaskSpecialWin ( xge_widget *wdg )
{
  xge_quatrotballd *qball;
  trans3f          tr;

  qball = (xge_quatrotballd*)wdg->data0;
  Trans3dTof ( qball->tr, &tr );
  XSetForeground ( xgedisplay, xge_specialpixmapgc, 1 );
  XDrawArc ( xgedisplay, xge_specialpixmap, xge_specialpixmapgc,
             0, 0, 2*XGE_QUBALL_R+1, 2*XGE_QUBALL_R+1, 0, 360*64 );
  _xge_QuatRotBallDrawCircles ( XGE_QUBALL_R, XGE_QUBALL_R, XGE_QUBALL_R,
                                &tr, _xge_OutSpecialMaskPixels, NULL );
} /*_xge_QuatRotBalldMaskSpecialWin*/
#endif

static void _xge_QuatRotBalldOpenSpecial ( xge_widget *er )
{
  xge_quatrotballd *qball;
  short            xc, yc, D;

  qball = (xge_quatrotballd*)er->data0;
  _xge_special_widget->msgproc = xge_EmptyMsg;
  _xge_special_widget->redraw = _xge_QuatRotBalldDrawSpecial;
  _xge_special_widget->id = -1;
  xc = er->x+XGE_QBALLICON_W/2-1;
  yc = er->y+XGE_QBALLICON_W/2-1;
  D = 2*XGE_QUBALL_R+2;
  _xge_special_widget->w = min ( D, xge_MAX_SPECIAL_WIDTH );
  _xge_special_widget->h = min ( D, xge_MAX_SPECIAL_HEIGHT );
  _xge_special_widget->x = xc-XGE_QUBALL_R;
  _xge_special_widget->y = yc-XGE_QUBALL_R;
  _xge_special_widget->data0 = qball;
  _xge_special_widget->data1 = er;
  _xge_special_widget->xofs = _xge_special_widget->yofs = 0;
  _xge_special_widget->rpos = 0;
  _xge_special_widget->window_num = er->window_num;
  _xge_special_widget->state = xgestate_NOTHING;
  _xge_special_widget->next = _xge_special_widget->prev =
  _xge_special_widget->up = NULL;
#ifdef USE_XEXT_SHAPE
  if ( xge_use_specialwin && !xge_specialwin_in_use ) {
    _xge_QuatRotCompSpecialWinSizes ( _xge_special_widget );
    _xge_MapSpecialWin ( er->window_num,
                         _xge_QuatRotCompSpecialWinSizes,
                         _xge_QuatRotBalldMaskSpecialWin,
                         _xge_QuatRotBallDrawSpecialWin, er );
  }
#endif
  xge_AddPopup ( _xge_special_widget, false, false );
} /*_xge_QuatRotBalldOpenSpecial*/

static void _xge_QuatRotBalldCloseSpecial ( void )
{
#ifdef USE_XEXT_SHAPE
  if ( xge_specialwin_in_use ) {
    _xge_UnmapSpecialWin ();
  }
#endif
  xge_RemovePopup ( false );
  _xge_special_widget->msgproc = xge_EmptyMsg;
  _xge_special_widget->redraw = xge_DrawEmpty;
  _xge_special_widget->id = -1;
  _xge_special_widget->w = _xge_special_widget->h = 0;
  _xge_special_widget->x = _xge_special_widget->y = -1;
  _xge_special_widget->data0 = _xge_special_widget->data1 = NULL;
  _xge_special_widget->xofs = _xge_special_widget->yofs = 0;
  _xge_special_widget->rpos = 0;
  _xge_special_widget->window_num = -1;
  _xge_special_widget->state = xgestate_NOTHING;
  _xge_special_widget->next = _xge_special_widget->prev =
  _xge_special_widget->up = NULL;
} /*_xge_uatRotBalldCloseSpecial*/

boolean xge_QuatRotBalldMsg ( xge_widget *er, int msg, int key, short x, short y )
{
  xge_quatrotballd *qball;
  int              xc, yc, r, d, dx, dy;

  qball = (xge_quatrotballd*)er->data0;
  xc = er->x+XGE_QBALLICON_W/2-1;
  yc = er->y+XGE_QBALLICON_H/2-1;
  r = XGE_QBALLICON_W/2+4;
  switch ( er->state ) {
case xgestate_NOTHING:
    switch ( msg ) {
  case xgemsg_ENTERING:
      goto setup_cursor;
  case xgemsg_EXITING:
      xge_SetCurrentWindowCursor ( xgeCURSOR_DEFAULT );
      return true;
  case xgemsg_MMOVE:
      d = (x-xc)*(x-xc) + (y-yc)*(y-yc);
      if ( d <= r*r )
        goto setup_cursor;
      else
        xge_SetCurrentWindowCursor ( xgeCURSOR_DEFAULT );
      return true;
  case xgemsg_MCLICK:
      d = (x-xc)*(x-xc) + (y-yc)*(y-yc);
      if ( d <= r*r ) {
        if ( key & (xgemouse_WHEELFW_CHANGE | xgemouse_WHEELBK_CHANGE) ) {
          if ( xge_QuatRotBalldRotateWheel ( er, key ) ) {
            xge_callback ( er, xgemsg_QUATROTBALL_COMMAND, 0, x, y );
            xge_SetClipping ( er );
            er->redraw ( er, true );
          }
        }
        else if ( (key & xgemouse_LBUTTON_DOWN) &&
                  (key & xgemouse_LBUTTON_CHANGE) ) {
          er->state = xgestate_QUATROT_TURNING1;
          goto enter_turning_state;
        }
        else if ( (key & xgemouse_RBUTTON_DOWN) &&
                  (key & xgemouse_RBUTTON_CHANGE) ) {
          er->state = xgestate_QUATROT_TURNING2;
enter_turning_state:
          qball->axis_ok = false;
          dx = xc +
               (int)(((double)(x-xc)*XGE_QUBALL_R)/(float)(XGE_QBALLICON_W/2-5)+0.5);
          dy = yc +
               (int)(((double)(y-yc)*XGE_QUBALL_R)/(float)(XGE_QBALLICON_H/2-5)+0.5);
          XWarpPointer ( xgedisplay, None, xgewindow, 0, 0, 0, 0, dx, dy );
          xge_xx = dx;
          xge_yy = dy;
          _xge_QuatRotBalldOpenSpecial ( er );
          xge_GrabFocus ( er, true );
          goto redraw_special;
        }
      }
      return true;
  case xgemsg_KEY:
      goto process_key;
  case xgemsg_SPECIAL_KEY:
      goto toggle_insert;
  default:
      return false;
    }

case xgestate_QUATROT_TURNING1:
    switch ( msg ) {
  case xgemsg_MMOVE:
      if ( !(key & xgemouse_LBUTTON_DOWN) )
        goto exit_turning_state;
      if ( key & xgemouse_RBUTTON_DOWN ) {
        er->state = xgestate_QUATROT_TURNING3;
        qball->axis_ok = false;
      }
      goto rotate_it;
  case xgemsg_MCLICK:
      if ( !(key & xgemouse_LBUTTON_DOWN) )
        goto exit_turning_state;
      else if ( key & (xgemouse_WHEELFW_CHANGE | xgemouse_WHEELBK_CHANGE) ) {
        if ( xge_QuatRotBalldRotateWheel ( er, key ) )
          goto turn_and_redraw_special;
      }
      else if ( key & xgemouse_RBUTTON_DOWN ) {
        er->state = xgestate_QUATROT_TURNING3;
        qball->axis_ok = false;
      }
      return true;
  case xgemsg_KEY:
      goto process_key;
  case xgemsg_SPECIAL_KEY:
      goto toggle_insert;
  default:
      return false;
    }

case xgestate_QUATROT_TURNING3:
    switch ( msg ) {
  case xgemsg_MMOVE:
      if ( !(key & xgemouse_LBUTTON_DOWN) )
        er->state = xgestate_QUATROT_TURNING2;
      else if ( !(key & xgemouse_RBUTTON_DOWN) )
        er->state = xgestate_QUATROT_TURNING1;
      goto rotate_it;
  case xgemsg_MCLICK:
      if ( !(key & xgemouse_LBUTTON_DOWN) )
        er->state = xgestate_QUATROT_TURNING2;
      else if ( !(key & xgemouse_RBUTTON_DOWN) )
        er->state = xgestate_QUATROT_TURNING1;
      return true;
  case xgemsg_KEY:
      goto process_key;
  case xgemsg_SPECIAL_KEY:
      goto toggle_insert;
  default:
      return false;
    }

case xgestate_QUATROT_TURNING2:
    switch ( msg ) {
  case xgemsg_MMOVE:
      if ( !(key & xgemouse_RBUTTON_DOWN) )
        goto exit_turning_state;
      if ( key & xgemouse_LBUTTON_DOWN ) {
        er->state = xgestate_QUATROT_TURNING3;
        qball->axis_ok = false;
      }
rotate_it:
      if ( xge_QuatRotBalldRotateIt ( er, x, y ) )
        goto turn_and_redraw_special;
      return true;
  case xgemsg_MCLICK:
      if ( (key & xgemouse_LBUTTON_DOWN) ) {
        er->state = xgestate_QUATROT_TURNING3;
        qball->axis_ok = false;
      }
      else if ( !(key & xgemouse_RBUTTON_DOWN) ) {
exit_turning_state:
        _xge_QuatRotBalldCloseSpecial ();
        er->state = xgestate_NOTHING;
        xge_ReleaseFocus ( er );
        goto redraw_special;
      }
      else if ( key & (xgemouse_WHEELFW_CHANGE | xgemouse_WHEELBK_CHANGE) ) {
        if ( xge_QuatRotBalldRotateWheel ( er, key ) )
          goto turn_and_redraw_special;
      }
      return true;
  case xgemsg_KEY:
process_key:
      switch ( key ) {
    case 'r':  case 'R':
        SetQuaterniond ( qball->q, 1.0, 0.0, 0.0, 0.0 );
        IdentTrans3d ( qball->tr );
turn_and_redraw_special:
        xge_callback ( er, xgemsg_QUATROTBALL_COMMAND, 0, x, y );
redraw_special:
#ifdef USE_XEXT_SHAPE
        if ( xge_specialwin_in_use )
          _xge_RemaskSpecialWin ();
#endif
        xge_Redraw ();
        return true;
    default:
        return false;
      }
  case xgemsg_SPECIAL_KEY:
toggle_insert:
      switch ( key ) {
    case 0xFF63:  /* insert - notebook Toshiba */
    case 0xFF9E:  /* insert - Vobis, Dell */
        qball->insert = !qball->insert;
setup_cursor:
        if ( qball->insert )
          xge_SetCurrentWindowCursor ( xgeCURSOR_ARROW );
        else
          xge_SetCurrentWindowCursor ( xgeCURSOR_DEFAULT );
        return true;
    default:
        return false;
      }
  default:
      return false;
    }

default:
    return false;
  }
} /*xge_QuatRotBalldMsg*/

xge_widget *xge_NewQuatRotBalld ( char window_num, xge_widget *prev, int id,
                                  short w, short h, short x, short y,
                                  xge_quatrotballd *qball, quaterniond *q,
                                  trans3d *tr, char *title )
{
  if ( !_xge_InitCanvas () )
    return NULL;
#ifdef USE_XEXT_SHAPE
  _xge_CreateSpecialWin ();
#endif
  if ( w < XGE_QBALLICON_W ) w = XGE_QBALLICON_W;
  if ( h < XGE_QBALLICON_H ) h = XGE_QBALLICON_H;
  qball->q = q;
  qball->tr = tr;
  SetQuaterniond ( qball->q, 1.0, 0.0, 0.0, 0.0 );
  IdentTrans3d ( qball->tr );
  qball->insert = qball->axis_ok = false;
  return (qball->er = xge_NewWidget ( window_num, prev, id, w, h, x, y,
                        qball, title, xge_QuatRotBalldMsg, xge_DrawQuatRotBalld ));
} /*xge_NewQuatRotBalld*/

