
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"

#include "xgedit.h"
#include "xgeprivate.h"

#define xgeid_FILEMENU_SAVE_DIRTEXT      0
#define xgeid_FILEMENU_SAVE_SW_HIDDEN    1
#define xgeid_FILEMENU_SAVE_FILE         2
#define xgeid_FILEMENU_SAVE_CANCEL       3
#define xgeid_FILEMENU_SAVE_DIRLIST      4
#define xgeid_FILEMENU_SAVE_FILELIST     5
#define xgeid_FILEMENU_SAVE_EDITFILENAME 6

/* ///////////////////////////////////////////////////////////////////////// */
static boolean _xge_SaveFileProc ( xge_widget *er,
                                   int msg, int key, short x, short y )
{
  xge_widget   *up;
  xge_filemenu *fm;

  up = er->up;
  fm = (xge_filemenu*)up->data2;
  switch ( msg ) {
case xgemsg_BUTTON_COMMAND:
    switch ( er->id ) {
  case xgeid_FILEMENU_SAVE_FILE:
      if ( xge_FilenameCorrect ( fm->filename ) )
        xge_callback ( up, xgemsg_FILEMENU_SAVEFILE, 0, 0, 0 );
      return true;
  case xgeid_FILEMENU_SAVE_CANCEL:
      xge_callback ( up, xgemsg_FILEMENU_CANCEL, 1, 0, 0 );
      xge_RemovePopup ( true );
      return true;
  default:
      return false;
    }

case xgemsg_SWITCH_COMMAND:
    switch ( er->id ) {
  case xgeid_FILEMENU_SAVE_SW_HIDDEN:
      return true;
  default:
      return false;
    }

case xgemsg_TEXT_WIDGET_CLICK:
    switch ( er->id ) {
  case xgeid_FILEMENU_SAVE_DIRTEXT:
      xge_ChangeDirAlt ( fm, x, false );
      xge_SetClipping ( fm->er2 );
      fm->er2->redraw ( fm->er2, true );
      return true;
  default:
      return false;
    }

case xgemsg_LISTBOX_ITEM_PICK:
    switch ( er->id ) {
  case xgeid_FILEMENU_SAVE_DIRLIST:
      xge_ChangeDir ( fm, false );
      xge_SetClipping ( fm->er2 );
      fm->er2->redraw ( fm->er2, true );
      return true;
  case xgeid_FILEMENU_SAVE_FILELIST:
      if ( xge_GetCurrentListBoxString ( &fm->filelist2, fm->filename ) ) {
        fm->lastfileitem2 = fm->filelist2.currentitem;
        fm->filename_editor.start = fm->filename_editor.pos = 0;
        xge_SetClipping ( fm->er2 );
        fm->er2->redraw ( fm->er2, true );
      }
      return true;
  default:
      return false;
    }

case xgemsg_TEXT_EDIT_VERIFY:
    switch ( er->id ) {
  case xgeid_FILEMENU_SAVE_EDITFILENAME:
      return xge_FilenameCorrect ( fm->filename );
  default:
      return 0;
    }

case xgemsg_TEXT_EDIT_ENTER:
    switch ( er->id ) {
  case xgeid_FILEMENU_SAVE_EDITFILENAME:
      return 1;
  default:
      return 0;
    }

case xgemsg_TEXT_EDIT_ESCAPE:
    switch ( er->id ) {
  case xgeid_FILEMENU_SAVE_EDITFILENAME:
      return 1;
  default:
      return 0;
    }

default:
    return false;
  }
} /*_xge_SaveFileProc*/

boolean xge_AddSaveFilePopup ( xge_filemenu *fm )
{
  xge_SetCurrentDir ( fm, false );
  xge_AddPopup ( fm->er2, true, true );
  return true;
} /*xge_AddSaveFilePopup*/

xge_widget *xge_InitSaveFileMenu ( char window_num, int id,
                                   short w, short h, short x, short y,
                                   xge_filemenu *fm, xge_widget *optw,
                                   char *fullpath, char *shortpath,
                                   char *prevpath, char *filter,
                                   char *filename )
{
  xge_widget *mw, *menu;

  if ( w < 400 ) w = 400;
  if ( h < 242 ) h = 232;
  fm->fullpath      = fullpath;
  fm->shortpath     = shortpath;
  fm->prevpath      = prevpath;
  fm->filter        = filter;
  fm->filename      = filename;
  fm->showhidden    = false;
  fm->lastfileitem2 = 0;
  mw = xge_NewTextWidget ( window_num, optw, xgeid_FILEMENU_SAVE_DIRTEXT,
                           380, 16, x+10, y+10, shortpath );
  mw = xge_NewSwitch ( window_num, mw, xgeid_FILEMENU_SAVE_SW_HIDDEN,
                       54, 16, x+10, y+30, xgetxt_hidden, &fm->showhidden );
  mw = xge_NewButton ( window_num, mw, xgeid_FILEMENU_SAVE_FILE,
                       58, 19, x+91, y+h-30, xgetxt_Save );
  mw = xge_NewButton ( window_num, mw, xgeid_FILEMENU_SAVE_CANCEL,
                       58, 19, x+251, y+h-30, xgetxt_Cancel );
  mw = xge_NewListBox ( window_num, mw, xgeid_FILEMENU_SAVE_DIRLIST,
                        180, 131, x+10, y+60, &fm->dirlist2 );
  mw = xge_NewListBox ( window_num, mw, xgeid_FILEMENU_SAVE_FILELIST,
                        180, 131, x+200+10, y+60, &fm->filelist2 );
  mw = xge_NewTextWidget ( window_num, mw, 0,
                           42, 16, x+200-44+10, y+30, xgetxt_save_as );
  mw = xge_NewStringEd ( window_num, mw, xgeid_FILEMENU_SAVE_EDITFILENAME,
                         180, 19, x+200+10, y+30, XGE_MAX_FILENAME_LENGTH,
                         fm->filename, &fm->filename_editor );

  fm->er2 = menu = xge_NewFMenu ( window_num, NULL, id, w, h, x, y, mw );
  menu->data2 = fm;
  menu->libmsgproc = _xge_SaveFileProc;
  menu->msgproc = xge_PopupMenuMsg;
  return fm->er2;
} /*xge_InitSaveFileMenu*/

