
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2010, 2016                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "xgedit.h"
#include "xgeprivate.h"


#define STEPS 64
void xge_DrawSlidebardRGB ( xge_widget *er, boolean onscreen )
{
  int           x, x1, steps, i;
  double        *slipos, a0, a1, t, *h, hsv[3], rgb[3];
  xgecolour_int c0, c1;

  slipos = er->data0;
  x = er->x+1;
  if ( er->state == xgestate_MOVINGSLIDE ) {
    a0 = 0.4;  a1 = 0.8;
  }
  else {
    a0 = 1.0;  a1 = 0.6;
  }
  steps = er->w > 2*STEPS ? STEPS : er->w/2;
  switch ( er->id & 0x07 ) {
case xgeid_CM_SL_RED:
    for ( i = 1; i <= steps; i++ ) {
      t = (double)i/(double)steps;
      c0 = xge_PixelColourf ( a0*t, 0.0, 0.0 );
      c1 = xge_PixelColourf ( a1*t, 0.0, 0.0 );
      x1 = er->x + 1 + ((er->w-2)*i)/steps;
      xge_DrawVShadedRect ( x1-x, er->h-2, x, er->y+1, c0, c1, er->h-2 );
      x = x1;
    }
    break;
case xgeid_CM_SL_GREEN:
    for ( i = 1; i <= steps; i++ ) {
      t = (double)i/(double)steps;
      c0 = xge_PixelColourf ( 0.0, a0*t, 0.0 );
      c1 = xge_PixelColourf ( 0.0, a1*t, 0.0 );
      x1 = er->x + 1 + ((er->w-2)*i)/steps;
      xge_DrawVShadedRect ( x1-x, er->h-2, x, er->y+1, c0, c1, er->h-2 );
      x = x1;
    }
    break;
case xgeid_CM_SL_BLUE:
    for ( i = 1; i <= steps; i++ ) {
      t = (double)i/(double)steps;
      c0 = xge_PixelColourf ( 0.0, 0.0, a0*t );
      c1 = xge_PixelColourf ( 0.0, 0.0, a1*t );
      x1 = er->x + 1 + ((er->w-2)*i)/steps;
      xge_DrawVShadedRect ( x1-x, er->h-2, x, er->y+1, c0, c1, er->h-2 );
      x = x1;
    }
    break;
case xgeid_CM_SL_SAT:
    if ( !(h = (double*)er->data1) )
      break;
    hsv[0] = *h;  hsv[1] = hsv[2] = 1.0;
    pkv_HSVtoRGBd ( hsv, rgb );
    for ( i = 1; i <= steps; i++ ) {
      t = (double)i/(double)steps;
      c0 = xge_PixelColourf ( a0*(t*rgb[0]+1.0-t), a0*(t*rgb[1]+1.0-t), a0*(t*rgb[2]+1.0-t) );
      c1 = xge_PixelColourf ( a1*(t*rgb[0]+1.0-t), a1*(t*rgb[1]+1.0-t), a1*(t*rgb[2]+1.0-t) );
      x1 = er->x + 1 + ((er->w-2)*i)/steps;
      xge_DrawVShadedRect ( x1-x, er->h-2, x, er->y+1, c0, c1, er->h-2 );
      x = x1;
    }
    break;
case xgeid_CM_SL_VALUE:
default:
    if ( !(h = (double*)er->data1) )
      break;
    hsv[0] = *h;  hsv[1] = hsv[2] = 1.0;
    pkv_HSVtoRGBd ( hsv, rgb );
    for ( i = 1; i <= steps; i++ ) {
      t = (double)i/(double)steps;
      c0 = xge_PixelColourf ( a0*t*rgb[0], a0*t*rgb[1], a0*t*rgb[2] );
      c1 = xge_PixelColourf ( a1*t*rgb[0], a1*t*rgb[1], a1*t*rgb[2] );
      x1 = er->x + 1 + ((er->w-2)*i)/steps;
      xge_DrawVShadedRect ( x1-x, er->h-2, x, er->y+1, c0, c1, er->h-2 );
      x = x1;
    }
    break;
  }
  xgeSetForeground ( xgec_WIDGET_FRAME );
  xgeDrawRectangle ( er->w-1, er->h-1, er->x, er->y );
  x = er->x + 2 + (int)((*slipos)*(double)(er->w - 10));
  xgeSetForeground ( xgec_WIDGET_FOREGROUND );
  xgeFillRectangle ( 6, 6, x, er->y+2 );
  if ( onscreen )
    xgeCopyRectOnScreen ( er->w, er->h, er->x, er->y );
} /*xge_DrawSlidebardRGB*/

xge_widget *xge_NewSlidebardRGB ( char window_num, xge_widget *prev, int id,
                                  short w, short h, short x, short y,
                                  double *data, double *hue )
{
  return xge_NewWidget ( window_num, prev, id, w, h, x, y, data, hue,
                         xge_SlidebardMsg, xge_DrawSlidebardRGB );
} /*xge_NewSlidebardRGB*/

