
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <GL/gl.h>  
#include <GL/glu.h>
#include <GL/glx.h>

#include "xgledit.h"
#include "xgeprivate.h"

/* ///////////////////////////////////////////////////////////////////////// */
/* The procedures in this file set up projection matrices, using formulae  */
/* employed by the deprecated OpenGL glFrustum and glOrtho procedures      */
/* It is possible to obtain at the same time their inverses (or only the   */
/* inverses. No OpenGL calls are present here, it is the matter of the     */
/* application to upload the matrices to appropriate uniform variables.    */
/* The coefficients are stored in the arrays in the column (OpenGL default */
/* order). */

/* perspective projection */
void xgle_Frustumd ( double left, double right, double bottom, double top,
                     double near, double far, GLdouble m[16], GLdouble im[16] )
{
  double rl, tb, nf, nn;

  rl = right-left;
  tb = top-bottom;
  nf = near-far;
  nn = near+near;
  if ( m ) {
    memset ( m, 0, 16*sizeof(GLdouble) );
    m[0] = nn/rl;           m[8] = (right+left)/rl;
    m[5] = nn/tb;           m[9] = (top+bottom)/tb;
    m[10] = (far+near)/nf;  m[14] = 2.0*far*near/nf;
    m[11] = -1.0;
  }
  if ( im ) {
    memset ( im, 0, 16*sizeof(GLdouble) );
    im[0] = rl/nn;  im[12] = (right+left)/nn;
    im[5] = tb/nn;  im[13] = (top+bottom)/nn;
    im[14] = -1.0;
    im[11] = 0.5*nf/(far*near);
    im[15] = 0.5*(far+near)/(far*near);
  }
} /*xgle_Frustumd*/

/* parallel orthogonal projection */
void xgle_Orthod ( double left, double right, double bottom, double top,
                   double near, double far, GLdouble m[16], GLdouble im[16] )
{
  double rl, tb, fn;

  rl = right-left;
  tb = top-bottom;
  fn = far-near;
  if ( m ) {
    memset ( m, 0, 16*sizeof(GLdouble) );
    m[0] = 2.0/rl;   m[12] = -(right+left)/rl;
    m[5] = 2.0/tb;   m[13] = -(top+bottom)/tb;
    m[10] = -2.0/fn;  m[14] = (far+near)/fn;
    m[15] = 1.0;
  }
  if ( im ) {
    memset ( im, 0, 16*sizeof(GLdouble) );
    im[0] = 0.5*rl;    im[12] = 0.5*(right+left);
    im[5] = 0.5*tb;    im[13] = 0.5*(top+bottom);
    im[10] = -0.5*fn;  im[14] = 0.5*(far+near);
    im[15] = 1.0;
  }
} /*xgle_Orthod*/

