
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "xgledit.h"
#include "xgeprivate.h"

GLuint xgle_CompileShaderFiles ( GLenum shader_type,
                                 int nfiles, const char *const*filenames )
{
  void   *sp;
  GLuint shader_id;
  FILE   *f;
  int    i;
  GLint  *fsize, totalsize;
  char   *src, **srclines;

  sp = pkv_GetScratchMemTop ();
  shader_id = 0;
  if ( nfiles > 0 ) {
    if ( !(fsize = pkv_GetScratchMem ( nfiles*sizeof(GLint) )) ||
         !(srclines = pkv_GetScratchMem ( nfiles*sizeof(char*) )) )
      goto way_out;
        /* find the necessary buffer size */
    for ( i = 0, totalsize = nfiles;  i < nfiles;  i++ ) {
      if ( !(f = fopen ( filenames[i], "rb" )) )
        goto way_out;
      fseek ( f, 0, SEEK_END );
      fsize[i] = ftell ( f );
      totalsize += fsize[i];
      fclose ( f );
    }
        /* read in the source files - let's hope they are not changed */
        /* in the meantime */
    if ( !(src = pkv_GetScratchMem ( totalsize )) )
      goto way_out;
    for ( i = 0, totalsize = 0;  i < nfiles;  i++ ) {
      if ( !(f = fopen ( filenames[i], "rb" )) )
        goto way_out;
      srclines[i] = &src[totalsize];
      if ( fread ( srclines[i], sizeof(char), fsize[i], f ) != fsize[i] )
        goto way_out;
      srclines[i][fsize[i]] = 0;
      totalsize += fsize[i]+1;
      fclose ( f );
    }
    shader_id = xgle_CompileShaderStrings ( shader_type, nfiles,
                                            (const GLchar**)srclines );
  }

way_out:
  pkv_SetScratchMemTop ( sp );
  return shader_id;
} /*xgle_CompileShaderFiles*/

